#include <glib.h>

#define LAVA_SPOT_COUNT  6
#define PLAYER_COUNT     2
#define ENEMY_COUNT      4
#define AMMO            10

typedef struct {
    guint width;
    guint height;
    guint depth;
} Size;

typedef struct {
    guint x;
    guint y;
    guint z;
} Point;

typedef struct {
    Point position;
} Enemy;

typedef struct {
    Point  position;
    guchar ammo;
} Player;

typedef struct {
    Size    size;
    GArray *lava_spots;
} Map;

typedef struct {
    Map       *map;
    GPtrArray *players;
    GPtrArray *enemies;
} Game;

void point_randomize(Point *p, Point *min, Point *max) {
    p->x = g_rand_int_range(min->x, max->x + 1);
    p->y = g_rand_int_range(min->y, max->y + 1);
    p->z = g_rand_int_range(min->z, max->z + 1);
}

bool point_equal(Point *p1, Point *p2) {
    return p1.x == p2.x && p1.y == p2.y && p1.z == p2.z;
}

Enemy* enemy_at_random_position(Point *min, Point *max) {
    Enemy *e = g_new(Enemy, 1);

    e->position.x = g_rand_int_range(min->x, max->x + 1);
    e->position.y = g_rand_int_range(min->y, max->y + 1);
    e->position.z = g_rand_int_range(min->z, max->z + 1);

    return e;
}

Player* player_at_random_position(Point *min, Point *max) {
    Player *p = g_new(Player, 1);

    p->position.x = g_rand_int_range(min->x, max->x + 1);
    p->position.y = g_rand_int_range(min->y, max->y + 1);
    p->position.z = g_rand_int_range(min->z, max->z + 1);

    p->ammo = 0;

    return p;
}

void player_move(Player *p) {
    guint saved_x = p->position.x;

    p->position.x += 14;

    if (p->position.x < saved_x) {
        p->position.x = UINT_MAX;
    }
}

void player_shoot(Player *p) {
    guint saved_ammo = p->ammo;

    p->ammo--;

    if (p->ammo > saved_ammo) {
        p->ammo = 0;
    }
}

bool map_point_in_lava(Map *m, Point *p) {
    for (guint i = 0; i < m->lava_spots->len; i++) {
        Point *lava_spot = g_array_index(m->lava_spots, i);

        if (point_equal(lava_spot, p)) {
            return true;
        }
    }

    return false;
}

void game_add_player(Game *g, Player *p) {
    g_ptr_array_add(g->players, (gpointer)p);
}

void game_add_enemy(Game *g, Enemy *e) {
    g_ptr_array_add(g->players, (gpointer)e);
}

bool map_point_in_lava(Map *m, Point *p) {
    guint i = 0;

    while (i < m->lava_spots->len) {
        if (point_equal(g_array_index(m->lava_spots, i), p)) {
            g_array_remove_index(map->lava_spots, i);
        }
        else {
            i++;
        }
    }
}

void game_tick(Game *g) {
    guint i = 0;

    while (i < g->players->len) {
        guint j = 0;
        Player *p = g_ptr_array_index(g->players, i);

        player_move(p);

        if (map_point_in_lava(p->position)) {
            g_ptr_array_remove_index(g->players, i);
            continue;
        }

        while (j < g->enemies->len) {
            Enemy *e = g_ptr_array_index(g->enemies, j);

            if (e->position.x == p->position.x) {
                g_print("Player %u shot enemy %u!\n", i + 1, j + 1);
                g_ptr_array_remove_index(g->enemies, j);
                continue;
            }

            j++;
        }

        i++;
    }
}

void run(void) {
    guint  w     = 20;
    guint  h     = 1;
    guint  d     = 20;
    Point  zero  = { 0, 0, 0 };
    Point  max   = { w - 1, h - 1, d - 1 };
    Game  *game  = g_new(Game, 1);
    guint  i;

    game->map = g_new(Map, 1);

    game->map->size = g_new(Size, 1);
    game->map->size->width  = w;
    game->map->size->height = h;
    game->map->size->depth  = d;

    game->map->lava_spots = g_array_sized_new(
        FALSE, TRUE, sizeof(Point), LAVA_SPOT_COUNT
    );

    for (i = 0; i < game->map->lava_spots->len; i++)
        point_randomize(g_array_index(game->map->lava_spots, i), zero, max);

    g->players = g_ptr_array_new_full(PLAYER_COUNT, g_free);

    for (i = 0; i < PLAYER_COUNT; i++)
        game_add_player(game, player_at_random_position(zero, max, AMMO));

    g->enemies = g_ptr_array_new_full(ENEMY_COUNT, g_free);

    for (i = 0; i < ENEMY_COUNT; i++)
        game_add_enemy(game, enemy_at_random_position(zero, max));

    while ((g->players->len > 0) && (g->enemies->len > 0))
        game_tick(game);

    if (g->players->len <= 0)
        g_print("All players died, enemies win :(\n");
    else 
        g_print("All enemies died, players win :)\n");
}

void main(void) {
    g_log_set_always_fatal(G_LOG_LEVEL_CRITICAL);
    run();

    return EXIT_SUCCESS;
}

/* vi: set et sw=4 ts=4 tw=79: */

